@@ -4,8 +4,8 @@
 package netlink_test
 
 import (
+	"errors"
 	"fmt"
-	"math"
 	"math/rand"
 	"net"
 	"os"
@@ -16,11 +16,10 @@ import (
 	"time"
 
 	"github.com/google/go-cmp/cmp"
-	"github.com/jsimonetti/rtnetlink"
+	"github.com/google/go-cmp/cmp/cmpopts"
 	"github.com/mdlayher/netlink"
 	"github.com/mdlayher/netlink/nlenc"
 	"golang.org/x/net/bpf"
-	"golang.org/x/net/nettest"
 	"golang.org/x/sys/unix"
 )
 
@@ -571,26 +570,6 @@ func testBPFProgram(allowSequence uint32) []bpf.Instruction {
 	}
 }
 
-func TestIntegrationConnMulticast(t *testing.T) {
-	skipUnprivileged(t)
-
-	c, done := rtnlDial(t, 0)
-	defer done()
-
-	// Create an interface to trigger a notification, and remove it at the end
-	// of the test.
-	const ifName = "nltest0"
-	defer shell(t, "ip", "link", "del", ifName)
-
-	ifi := rtnlReceive(t, c, func() {
-		shell(t, "ip", "tuntap", "add", ifName, "mode", "tun")
-	})
-
-	if diff := cmp.Diff(ifName, ifi); diff != "" {
-		t.Fatalf("unexpected interface name (-want +got):\n%s", diff)
-	}
-}
-
 func TestIntegrationConnExplicitPID(t *testing.T) {
 	t.Parallel()
 
@@ -598,7 +577,7 @@ func TestIntegrationConnExplicitPID(t *testing.T) {
 	// PID will be used in messages that are sent to and received from the
 	// kernel.
 	rng := rand.New(rand.NewSource(time.Now().UnixNano()))
-	pid := uint32(rng.Intn(math.MaxUint32))
+	pid := rng.Uint32()
 
 	c, err := netlink.Dial(unix.NETLINK_GENERIC, &netlink.Config{PID: pid})
 	if err != nil {
@@ -720,256 +699,174 @@ func TestIntegrationConnExecuteTimeout(t *testing.T) {
 	mustBeTimeoutNetError(t, err)
 }
 
-func TestIntegrationRTNetlinkStrictCheckExtendedAcknowledge(t *testing.T) {
-	c, err := netlink.Dial(unix.NETLINK_ROUTE, nil)
-	if err != nil {
-		t.Fatalf("failed to open rtnetlink socket: %s", err)
-	}
-	defer c.Close()
-
-	// Turn on extended acknowledgements and strict checking so rtnetlink
-	// reports detailed error information regarding our invalid dump request.
-	setStrictCheck(t, c)
-	if err := c.SetOption(netlink.ExtendedAcknowledge, true); err != nil {
-		t.Fatalf("failed to set extended acknowledge option: %v", err)
-	}
-
-	// The kernel will complain that this field isn't valid for a filtered dump
-	// request.
-	b, err := (&rtnetlink.RouteMessage{SrcLength: 1}).MarshalBinary()
-	if err != nil {
-		t.Fatalf("failed to marshal request: %v", err)
-	}
-
-	_, err = c.Execute(netlink.Message{
-		Header: netlink.Header{
-			Type:  unix.RTM_GETROUTE,
-			Flags: netlink.Request | netlink.Dump,
+func TestOpErrorUnwrapLinux(t *testing.T) {
+	tests := []struct {
+		name   string
+		err    error
+		target error
+		ok     bool
+	}{
+		{
+			name:   "ENOBUFS",
+			err:    unix.ENOBUFS,
+			target: os.ErrNotExist,
+		},
+		{
+			name: "OpError ENOBUFS",
+			err: &netlink.OpError{
+				Op:  "receive",
+				Err: unix.ENOBUFS,
+			},
+			target: os.ErrNotExist,
+		},
+		{
+			name: "OpError os.SyscallError ENOBUFS",
+			err: &netlink.OpError{
+				Op:  "receive",
+				Err: os.NewSyscallError("recvmsg", unix.ENOBUFS),
+			},
+			target: os.ErrNotExist,
+		},
+		{
+			name:   "ENOENT",
+			err:    unix.ENOENT,
+			target: os.ErrNotExist,
+			ok:     true,
+		},
+		{
+			name: "OpError ENOENT",
+			err: &netlink.OpError{
+				Op:  "receive",
+				Err: unix.ENOENT,
+			},
+			target: os.ErrNotExist,
+			ok:     true,
+		},
+		{
+			name: "OpError os.SyscallError ENOENT",
+			err: &netlink.OpError{
+				Op:  "receive",
+				Err: os.NewSyscallError("recvmsg", unix.ENOENT),
+			},
+			target: os.ErrNotExist,
+			ok:     true,
+		},
+		{
+			name: "OpError os.SyscallError EEXIST",
+			err: &netlink.OpError{
+				Op:  "receive",
+				Err: os.NewSyscallError("recvmsg", unix.EEXIST),
+			},
+			target: os.ErrExist,
+			ok:     true,
 		},
-		Data: b,
-	})
-
-	oerr, ok := err.(*netlink.OpError)
-	if !ok {
-		t.Fatalf("expected *netlink.OpError, but got: %T", err)
 	}
 
-	// Assume the message contents will be relatively static but don't hardcode
-	// offset just in case things change.
-
-	want := &netlink.OpError{
-		Op:      "receive",
-		Err:     unix.EINVAL,
-		Message: "Invalid values in header for FIB dump request",
-	}
-
-	if diff := cmp.Diff(want, oerr); diff != "" {
-		t.Fatalf("unexpected *netlink.OpError (-want +got):\n%s", diff)
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			got := errors.Is(tt.err, tt.target)
+			if diff := cmp.Diff(tt.ok, got); diff != "" {
+				t.Fatalf("unexpected result (-want +got):\n%s", diff)
+			}
+		})
 	}
 }
 
-func TestIntegrationRTNetlinkRouteManipulation(t *testing.T) {
-	skipUnprivileged(t)
+func TestIntegrationConnClosedConn(t *testing.T) {
+	t.Parallel()
 
-	c, err := netlink.Dial(unix.NETLINK_ROUTE, nil)
+	c, err := netlink.Dial(unix.NETLINK_GENERIC, nil)
 	if err != nil {
-		t.Fatalf("failed to open rtnetlink socket: %s", err)
-	}
-	defer c.Close()
-
-	// Required for kernel route dump filtering.
-	setStrictCheck(t, c)
-
-	lo, err := nettest.LoopbackInterface()
-	if err != nil {
-		t.Fatalf("failed to get loopback: %v", err)
+		t.Fatalf("failed to dial netlink: %v", err)
 	}
 
-	// Install synthetic routes in documentation ranges into a non-default table
-	// which we will later dump.
-	const (
-		table   = 100
-		ip4Mask = 32
-		ip6Mask = 128
-	)
+	// Close the connection immediately and ensure that future calls get EBADF.
+	if err := c.Close(); err != nil {
+		t.Fatalf("failed to close: %v", err)
+	}
 
-	var (
-		ip4 = &net.IPNet{
-			IP:   net.IPv4(192, 2, 2, 1),
-			Mask: net.CIDRMask(ip4Mask, ip4Mask),
-		}
-
-		ip6 = &net.IPNet{
-			IP:   net.ParseIP("2001:db8::1"),
-			Mask: net.CIDRMask(ip6Mask, ip6Mask),
-		}
-
-		want = []*net.IPNet{ip4, ip6}
-	)
-
-	rtmsgs := []rtnetlink.RouteMessage{
+	tests := []struct {
+		name string
+		fn   func() error
+	}{
 		{
-			Family:    unix.AF_INET,
-			DstLength: ip4Mask,
-			Protocol:  unix.RTPROT_STATIC,
-			Scope:     unix.RT_SCOPE_UNIVERSE,
-			Type:      unix.RTN_UNICAST,
-			Attributes: rtnetlink.RouteAttributes{
-				Dst:      ip4.IP,
-				OutIface: uint32(lo.Index),
-				Table:    table,
+			name: "receive",
+			fn: func() error {
+				_, err := c.Receive()
+				return err
 			},
 		},
 		{
-			Family:    unix.AF_INET6,
-			DstLength: ip6Mask,
-			Protocol:  unix.RTPROT_STATIC,
-			Scope:     unix.RT_SCOPE_UNIVERSE,
-			Type:      unix.RTN_UNICAST,
-			Attributes: rtnetlink.RouteAttributes{
-				Dst:      ip6.IP,
-				OutIface: uint32(lo.Index),
-				Table:    table,
+			name: "send",
+			fn: func() error {
+				_, err := c.Send(netlink.Message{})
+				return err
+			},
+		},
+		{
+			name: "set option",
+			fn: func() error {
+				return c.SetOption(netlink.ExtendedAcknowledge, true)
+			},
+		},
+		{
+			name: "syscall conn",
+			fn: func() error {
+				_, err := c.SyscallConn()
+				return err
 			},
 		},
 	}
 
-	// Verify we can send a batch of updates in one syscall.
-	var msgs []netlink.Message
-	for _, m := range rtmsgs {
-		b, err := m.MarshalBinary()
-		if err != nil {
-			t.Fatalf("failed to marshal: %v", err)
-		}
-
-		msgs = append(msgs, netlink.Message{
-			Header: netlink.Header{
-				Type:  unix.RTM_NEWROUTE,
-				Flags: netlink.Request | netlink.Create | netlink.Replace,
-			},
-			Data: b,
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			if diff := cmp.Diff(unix.EBADF, tt.fn(), cmpopts.EquateErrors()); diff != "" {
+				t.Fatalf("unexpected error (-want +got):\n%s", diff)
+			}
 		})
 	}
-
-	if _, err := c.SendMessages(msgs); err != nil {
-		t.Fatalf("failed to add routes: %v", err)
-	}
-
-	// Only dump routes from the specified table.
-	b, err := (&rtnetlink.RouteMessage{
-		Attributes: rtnetlink.RouteAttributes{Table: table},
-	}).MarshalBinary()
-	if err != nil {
-		t.Fatalf("failed to marshal request: %v", err)
-	}
-
-	routes, err := c.Execute(
-		netlink.Message{
-			Header: netlink.Header{
-				Type:  unix.RTM_GETROUTE,
-				Flags: netlink.Request | netlink.Dump,
-			},
-			Data: b,
-		},
-	)
-	if err != nil {
-		t.Fatalf("failed to dump routes: %v", err)
-	}
-
-	// Parse the routes back to Go structures.
-	got := make([]*net.IPNet, 0, len(routes))
-	for _, r := range routes {
-		var rtm rtnetlink.RouteMessage
-		if err := rtm.UnmarshalBinary(r.Data); err != nil {
-			t.Fatalf("failed to unmarshal route: %v", err)
-		}
-
-		got = append(got, &net.IPNet{
-			IP:   rtm.Attributes.Dst,
-			Mask: net.CIDRMask(int(rtm.DstLength), int(rtm.DstLength)),
-		})
-	}
-
-	// Now clear the routes and verify they're removed before ensuring we got
-	// the expected routes.
-	for i := range msgs {
-		msgs[i].Header.Type = unix.RTM_DELROUTE
-		msgs[i].Header.Flags = netlink.Request | netlink.Acknowledge
-	}
-
-	if _, err := c.SendMessages(msgs); err != nil {
-		t.Fatalf("failed to send: %v", err)
-	}
-	if _, err := c.Receive(); err != nil {
-		t.Fatalf("failed to receive: %v", err)
-	}
-
-	if diff := cmp.Diff(want, got); diff != "" {
-		t.Fatalf("unexpected routes (-want +got):\n%s", diff)
-	}
 }
 
-func TestIntegrationConnNetNSExplicit(t *testing.T) {
-	skipUnprivileged(t)
-
-	// Create a network namespace for use within this test.
-	const ns = "nltest0"
-	shell(t, "ip", "netns", "add", ns)
-	defer shell(t, "ip", "netns", "del", ns)
-
-	f, err := os.Open("/var/run/netns/" + ns)
+func TestIntegrationConnStrict(t *testing.T) {
+	c, err := netlink.Dial(unix.NETLINK_GENERIC, &netlink.Config{Strict: true})
 	if err != nil {
-		t.Fatalf("failed to open namespace file: %v", err)
+		if errors.Is(err, unix.ENOPROTOOPT) {
+			t.Skipf("skipping, strict options not supported by this kernel: %v", err)
+		}
+
+		t.Fatalf("failed to dial netlink: %v", err)
 	}
-	defer f.Close()
+	defer c.Close()
 
-	// Create a connection in each the host namespace and the new network
-	// namespace. We will use these to validate that a namespace was entered
-	// and that an interface creation notification was only visible to the
-	// connection within the namespace.
-	hostC, hostDone := rtnlDial(t, 0)
-	defer hostDone()
+	sc, err := c.SyscallConn()
+	if err != nil {
+		t.Fatalf("failed to open syscall conn: %v", err)
+	}
 
-	nsC, nsDone := rtnlDial(t, int(f.Fd()))
-	defer nsDone()
+	// Strict mode applies a series of socket options. Check each applied option
+	// and update the map to true if we found it set to true. Any options which
+	// were not applied as expected will result in the test failing.
+	opts := map[int]bool{
+		unix.NETLINK_EXT_ACK:        false,
+		unix.NETLINK_GET_STRICT_CHK: false,
+	}
 
-	var wg sync.WaitGroup
-	wg.Add(1)
-	defer wg.Wait()
-
-	go func() {
-		defer wg.Done()
-
-		_, err := hostC.Receive()
-		if err == nil {
-			panic("received netlink message in host namespace")
+	err = sc.Control(func(fd uintptr) {
+		for k := range opts {
+			// The kernel returns a non-zero value for true.
+			if v, err := unix.GetsockoptInt(int(fd), unix.SOL_NETLINK, k); err == nil && v != 0 {
+				opts[k] = true
+			}
 		}
-
-		// Timeout means we were interrupted, so return.
-		if nerr, ok := err.(net.Error); ok && nerr.Timeout() {
-			return
-		}
-
-		panicf("failed to receive in host namespace: %v", err)
-	}()
-
-	// Create a temporary interface within the new network namespace.
-	const ifName = "nltestns0"
-	defer shell(t, "ip", "netns", "exec", ns, "ip", "link", "del", ifName)
-
-	ifi := rtnlReceive(t, nsC, func() {
-		// Trigger a notification in the new namespace.
-		shell(t, "ip", "netns", "exec", ns, "ip", "tuntap", "add", ifName, "mode", "tun")
 	})
-
-	// And finally interrupt the host connection so it can exit its
-	// receive goroutine.
-	if err := hostC.SetDeadline(time.Unix(1, 0)); err != nil {
-		t.Fatalf("failed to interrupt host connection: %v", err)
+	if err != nil {
+		t.Fatalf("failed to call control: %v", err)
 	}
 
-	if diff := cmp.Diff(ifName, ifi); diff != "" {
-		t.Fatalf("unexpected interface name (-want +got):\n%s", diff)
+	for k, v := range opts {
+		if !v {
+			t.Errorf("socket option %d was not set to true", k)
+		}
 	}
 }
 
@@ -985,69 +882,6 @@ func mustBeTimeoutNetError(t *testing.T, err error) {
 	}
 }
 
-func rtnlDial(t *testing.T, netNS int) (*netlink.Conn, func()) {
-	t.Helper()
-
-	timer := time.AfterFunc(10*time.Second, func() {
-		panic("test took too long")
-	})
-
-	c, err := netlink.Dial(unix.NETLINK_ROUTE, &netlink.Config{
-		Groups: unix.RTMGRP_LINK,
-		NetNS:  netNS,
-	})
-	if err != nil {
-		t.Fatalf("failed to dial rtnetlink: %v", err)
-	}
-
-	return c, func() {
-		if err := c.Close(); err != nil {
-			t.Fatalf("failed to close rtnetlink connection: %v", err)
-		}
-
-		// Stop the timer to prevent a panic if other tests run for a long time.
-		timer.Stop()
-	}
-}
-
-func rtnlReceive(t *testing.T, c *netlink.Conn, do func()) string {
-	t.Helper()
-
-	// Receive messages in goroutine.
-	msgC := make(chan rtnetlink.LinkMessage)
-	go func() {
-		msgs, err := c.Receive()
-		if err != nil {
-			panicf("failed to receive rtnetlink messages: %v", err)
-		}
-
-		var rtmsg rtnetlink.LinkMessage
-		if err := rtmsg.UnmarshalBinary(msgs[0].Data); err != nil {
-			panicf("failed to unmarshal rtnetlink message: %v", err)
-		}
-
-		msgC <- rtmsg
-	}()
-
-	// Execute the function which will generate messages, and then wait for
-	// a message.
-	do()
-	m := <-msgC
-
-	return m.Attributes.Name
-}
-
-func setStrictCheck(t *testing.T, c *netlink.Conn) {
-	if err := c.SetOption(netlink.GetStrictCheck, true); err != nil {
-		// TODO(mdlayher): use errors.Is when we drop support for Go 1.12.
-		if err.(*netlink.OpError).Err == unix.ENOPROTOOPT {
-			t.Skipf("skipping, netlink strict checking is not supported on this kernel")
-		}
-
-		t.Fatalf("failed to set strict check option: %v", err)
-	}
-}
-
 func skipPrivileged(t *testing.T) {
 	u, err := user.Current()
 	if err != nil {
